"""Harmonic lattice vibrations."""

import numpy as np
from numpy.typing import NDArray

from ramannoodle.abstract import Dynamics
from ramannoodle.constants import RAMAN_TENSOR_CENTRAL_DIFFERENCE
from ramannoodle.abstract import PolarizabilityModel
from ramannoodle.spectrum._raman import PhononRamanSpectrum
from ramannoodle.exceptions import verify_ndarray_shape


class Phonons(Dynamics):
    r"""Harmonic lattice vibrations.

    A phonon can be represented by a wavenumber and corresponding atomic displacement.
    The wavenumbers are the eigenvalues of the system's dynamical matrix, while the
    atomic displacements are the eigenvectors of the dynamical matrix divided by the
    square root of the atomic masses.

    Parameters
    ----------
    ref_positions
        (fractional) Array with shape (N,3) where N is the number of atoms.
    wavenumbers
        (cm\ :sup:`-1`) Array with shape (M,).
    displacements
        (fractional) Array with shape (M,N,3).
    """

    def __init__(
        self,
        ref_positions: NDArray[np.float64],
        wavenumbers: NDArray[np.float64],
        displacements: NDArray[np.float64],
    ) -> None:
        verify_ndarray_shape("ref_positions", ref_positions, (None, 3))
        verify_ndarray_shape("wavenumbers", wavenumbers, (None,))
        verify_ndarray_shape(
            "displacements",
            displacements,
            (wavenumbers.size, ref_positions.shape[0], 3),
        )
        self._ref_positions = ref_positions
        self._wavenumbers: NDArray[np.float64] = wavenumbers
        self._displacements: NDArray[np.float64] = displacements

    @property
    def ref_positions(self) -> NDArray[np.float64]:
        r"""Get (a copy of) reference positions.

        Returns
        -------
        :
            (fractional) Array with shape (N,3) where N is the number of atoms.
        """
        return self._ref_positions.copy()

    @property
    def wavenumbers(self) -> NDArray[np.float64]:
        r"""Get (a copy of) wavenumbers.

        Returns
        -------
        :
            (cm\ :sup:`-1`) Array with shape (M,) where M is the number of phonons.
        """
        return self._wavenumbers.copy()

    @property
    def displacements(self) -> NDArray[np.float64]:
        """Get (a copy of) displacements.

        Returns
        -------
        :
            (fractional) Array with shape (M,N,3) where M is the number of phonons
            and N is the number of atoms.
        """
        return self._displacements.copy()

    def get_raman_spectrum(
        self, polarizability_model: PolarizabilityModel
    ) -> PhononRamanSpectrum:
        """Calculate a raman spectrum using a polarizability model.

        Parameters
        ----------
        polarizability_model
            Must be compatible with phonons.
        """
        raman_tensors = []
        for displacement in self._displacements:
            try:
                epsilon = displacement * RAMAN_TENSOR_CENTRAL_DIFFERENCE
                plus = polarizability_model.calc_polarizabilities(
                    np.array([self.ref_positions + epsilon])
                )[0]
                minus = polarizability_model.calc_polarizabilities(
                    np.array([self.ref_positions - epsilon])
                )[0]
            except ValueError as exc:
                raise ValueError(
                    "polarizability_model and phonons are incompatible"
                ) from exc
            raman_tensors.append((plus - minus) / RAMAN_TENSOR_CENTRAL_DIFFERENCE)

        return PhononRamanSpectrum(self._wavenumbers, np.array(raman_tensors))



"""Molecular dynamics trajectory."""

from collections.abc import Sequence
from typing import overload

import numpy as np
from numpy.typing import NDArray

from ramannoodle.abstract import Dynamics
from ramannoodle.abstract import PolarizabilityModel
from ramannoodle.exceptions import verify_ndarray_shape, get_type_error
from ramannoodle.spectrum._raman import MDRamanSpectrum
from ramannoodle.structure.utils import apply_pbc


class Trajectory(Dynamics, Sequence[NDArray[np.float64]]):
    r"""Molecular dynamics trajectory.

    A trajectory is represented as a positions time series (``positions_ts``) and a
    timestep.

    Parameters
    ----------
    positions_ts
        (fractional) Array with shape (S,N,3) where S in the number of
        configurations and N is the number of atoms.
    timestep
        (fs)

    """

    def __init__(
        self,
        positions_ts: NDArray[np.float64],
        timestep: float,
    ) -> None:
        verify_ndarray_shape("positions_ts", positions_ts, (None, None, 3))
        try:
            timestep = float(timestep)
        except TypeError as exc:
            raise get_type_error("timestep", timestep, "float") from exc
        if timestep <= 0:
            raise ValueError("timestep must be positive")

        self._positions_ts = apply_pbc(positions_ts)
        self._timestep = timestep

    @property
    def positions_ts(self) -> NDArray[np.float64]:
        r"""Get (a copy of) the positions time series.

        Returns
        -------
        :
            (fractional) Array with shape (S,N,3) where S in the number of
            configurations and N is the number of atoms.
        """
        return self._positions_ts.copy()

    @property
    def timestep(self) -> float:
        """Get timestep.

        Returns
        -------
        :
            (fs)
        """
        return self._timestep

    def get_raman_spectrum(
        self, polarizability_model: PolarizabilityModel
    ) -> MDRamanSpectrum:
        """Calculate a raman spectrum using a polarizability model.

        Parameters
        ----------
        polarizability_model
            Must be compatible with the trajectory.
        """
        try:
            polarizability_ts = polarizability_model.calc_polarizabilities(
                self._positions_ts
            )
        except ValueError as exc:
            raise ValueError(
                "polarizability_model and trajectory are incompatible"
            ) from exc

        return MDRamanSpectrum(polarizability_ts, self._timestep)

    def __len__(self) -> int:
        """Get trajectory length in timesteps."""
        return len(self._positions_ts)

    @overload
    def __getitem__(self, key: int) -> NDArray[np.float64]: ...

    @overload
    def __getitem__(self, key: slice) -> NDArray[np.float64]: ...

    def __getitem__(self, key: int | slice) -> NDArray[np.float64]:
        """Get positions (supports indexing and slicing)."""
        try:
            return self._positions_ts[key]
        except IndexError as exc:
            if "out of bounds" in str(exc):
                raise IndexError("trajectory index out of bounds") from exc
            raise exc



"""Functions for interacting with VASP XDATCAR files."""

from pathlib import Path

import numpy as np
from numpy.typing import NDArray

from ramannoodle.io._utils import verify_trajectory, pathify
from ramannoodle.exceptions import InvalidFileException
from ramannoodle.dynamics._trajectory import Trajectory
from ramannoodle.io.vasp.poscar import (
    _read_lattice,
    _read_atomic_symbols,
    _read_positions,
    _get_lattice_str,
    _get_symbols_str,
    _get_positions_str,
)


def read_positions_ts(
    filepath: str | Path,
) -> NDArray[np.float64]:
    """Read fractional positions time series from a VASP XDATCAR file.

    Parameters
    ----------
    filepath

    Returns
    -------
    :
        (fractional) Array with shape (S,N,3) where S is the number of configurations
        and N is the number of atoms.

    Raises
    ------
    FileNotFoundError
        File not found.
    InvalidFileException
        Invalid file.
    """
    filepath = pathify(filepath)
    positions_ts = []
    with open(filepath, "r", encoding="utf-8") as file:
        lattice = _read_lattice(file)
        atomic_symbols = _read_atomic_symbols(file)
        while True:
            try:
                positions = _read_positions(file, lattice, len(atomic_symbols))
                positions_ts.append(positions)
            except InvalidFileException as exc:
                if "missing first character in coordinate format:" in str(exc):
                    break
                raise exc
        return np.array(positions_ts)


def read_trajectory(
    filepath: str | Path,
    timestep: float,
) -> Trajectory:
    """Read trajectory from a VASP XDATCAR file.

    Timestep must be manually specified, as XDATCAR's do not contain the timestep.

    Parameters
    ----------
    filepath
    timestep
        (fs)

    Raises
    ------
    FileNotFoundError
        File not found.
    InvalidFileException
        Invalid file.
    """
    positions_ts = read_positions_ts(filepath)
    return Trajectory(positions_ts, timestep)


# pylint: disable=too-many-arguments,too-many-positional-arguments
def write_trajectory(
    lattice: NDArray[np.float64],
    atomic_numbers: list[int],
    positions_ts: NDArray[np.float64],
    filepath: str | Path,
    overwrite: bool = False,
    label: str = "XDATCAR written by ramannoodle",
) -> None:
    """Write trajectory to a VASP XDATCAR file.

    Parameters
    ----------
    lattice
        (Å) Array with shape (3,3).
    atomic_numbers
        List of length N where N is the number of atoms.
    positions_ts
        (fractional) Array with shape (S,N,3) where S is the number of
        configurations.
    filepath
    overwrite
        Overwrite the file if it exists.
    label
        XDATCAR label (first line).
    """
    verify_trajectory(lattice, atomic_numbers, positions_ts)
    filepath = pathify(filepath)

    open_mode = "w" if overwrite else "x"
    filepath = pathify(filepath)

    label_str = repr(label)[1:-1] + "\n"  # Raw string with quotes removed
    lattice_str = _get_lattice_str(lattice)
    symbols_str = _get_symbols_str(atomic_numbers)
    with open(filepath, open_mode, encoding="utf-8") as file:
        file.write(label_str)
        file.write("   1.00000000000000" + "\n")
        file.write(lattice_str)
        file.write(symbols_str)
        for index, positions in enumerate(positions_ts):
            config_label = f"Direct configuration= {index+1:>6}\n"
            positions_str = _get_positions_str(positions)
            file.write(config_label)
            file.write(positions_str)



"""Classes representing atomic motions, for example phonons and trajectories."""

# flake8: noqa: F401
from ramannoodle.dynamics._phonon import Phonons
from ramannoodle.dynamics._trajectory import Trajectory

__all__ = ["Phonons", "Trajectory"]



"""Functions for interacting with VASP OUTCAR files."""

from typing import TextIO
from pathlib import Path
import numpy as np
from numpy.typing import NDArray

from ramannoodle.io._utils import (
    _skip_file_until_line_contains,
    pathify,
    _read_polarizability_dataset,
)
from ramannoodle.exceptions import InvalidFileException, NoMatchingLineFoundException
from ramannoodle.constants import ATOMIC_WEIGHTS, ATOMIC_NUMBERS
from ramannoodle.exceptions import get_type_error, UserError
from ramannoodle.dynamics._phonon import Phonons
from ramannoodle.dynamics._trajectory import Trajectory
from ramannoodle.structure._reference import ReferenceStructure

try:
    from ramannoodle.dataset.torch._dataset import PolarizabilityDataset
except UserError:
    pass


# Utilities for OUTCAR. Warning: some of these functions partially read files.
def _get_atomic_symbol_from_potcar_line(line: str) -> str:
    """Read atomic symbol from a POTCAR line in a VASP OUTCAR file.

    Raises
    ------
    TypeError
    ValueError
    """
    try:
        symbol = line.split()[-2].split("_")[0]
    except AttributeError as exc:
        raise get_type_error("line", line, "str") from exc
    except IndexError as exc:
        raise ValueError(f"could not parse atomic symbol: {line}") from exc
    if symbol not in ATOMIC_NUMBERS:
        raise ValueError(f"unrecognized atomic symbol '{symbol}': {line}")
    return symbol


def _read_atomic_symbols(file: TextIO) -> list[str]:
    """Read atomic symbols from a VASP OUTCAR file.

    Raises
    ------
    InvalidFileException

    """
    # Get atom symbols first
    potcar_symbols: list[str] = []
    try:
        line = _skip_file_until_line_contains(file, "POTCAR:    ")
        potcar_symbols.append(_get_atomic_symbol_from_potcar_line(line))
        for line in file:
            if "POTCAR" not in line:
                break
            potcar_symbols.append(_get_atomic_symbol_from_potcar_line(line))

        # HACK: We read the next line and clean up as appropriate.
        # I wish the OUTCAR format was easier to parse, but alas, here we are.
        line = file.readline()
        if "VRHFIN" in line:
            potcar_symbols.pop()  # We read one too many!
    except NoMatchingLineFoundException as exc:
        raise InvalidFileException("POTCAR block not found") from exc
    except ValueError as exc:
        raise InvalidFileException(f"POTCAR block could not be parsed: {line}") from exc

    # Then get atom numbers
    try:
        line = _skip_file_until_line_contains(file, "ions per type")
        atomic_numbers = [int(item) for item in line.split()[4:]]

        atomic_symbols = []
        for symbol, number in zip(potcar_symbols, atomic_numbers):
            atomic_symbols += [symbol] * number
        return atomic_symbols
    except (NoMatchingLineFoundException, IndexError) as exc:
        raise InvalidFileException(
            f"ion number block could not be parsed: {line}"
        ) from exc


def _read_eigenvector(
    file: TextIO, num_atoms: int, lattice: NDArray[np.float64]
) -> NDArray[np.float64]:
    """Read the next available phonon eigenvector from a VASP OUTCAR file.

    Raises
    ------
    InvalidFileException
    """
    eigenvector: list[NDArray[np.float64]] = []
    try:
        for line in file:
            if len(eigenvector) == num_atoms:
                break
            if "X" in line:
                continue
            vector = [float(item) for item in line.split()[3:]]
            vector = vector @ np.linalg.inv(lattice)
            eigenvector.append(vector)
        return np.array(eigenvector)
    except (ValueError, IndexError) as exc:
        raise InvalidFileException(f"eigenvector could not be parsed: {line}") from exc


def _read_cart_positions(file: TextIO, num_atoms: int) -> NDArray[np.float64]:
    """Read atomic cartesian positions from a VASP OUTCAR file.

    Raises
    ------
    InvalidFileException
    """
    try:
        _ = _skip_file_until_line_contains(
            file, "position of ions in cartesian coordinates  (Angst):"
        )
    except NoMatchingLineFoundException as exc:
        raise InvalidFileException("cartesian positions not found") from exc
    cart_coordinates = []
    for _ in range(num_atoms):
        try:
            line = file.readline()
            cart_coordinates.append([float(item) for item in line.split()[0:3]])
        except (EOFError, ValueError, IndexError) as exc:
            raise InvalidFileException(
                f"cartesian positions could not be parsed: {line}"
            ) from exc

    return np.array(cart_coordinates)


def _read_positions(file: TextIO, num_atoms: int) -> NDArray[np.float64]:
    """Read atomic fractional positions from a VASP OUTCAR file.

    Raises
    ------
    InvalidFileException
    """
    try:
        _ = _skip_file_until_line_contains(
            file, "position of ions in fractional coordinates (direct lattice)"
        )
    except NoMatchingLineFoundException as exc:
        raise InvalidFileException("fractional positions not found") from exc

    positions = []
    for _ in range(num_atoms):
        try:
            line = file.readline()
            positions.append([float(item) for item in line.split()[0:3]])
        except (EOFError, ValueError, IndexError) as exc:
            raise InvalidFileException(
                f"fractional positions could not be parsed: {line}"
            ) from exc
    return np.array(positions)


def _read_polarizability(file: TextIO) -> NDArray[np.float64]:
    """Read polarizability from a VASP OUTCAR file.

    In actuality, we read the macroscopic dielectric tensor including local field
    effects.

    Raises
    ------
    InvalidFileException
    """
    try:
        _ = _skip_file_until_line_contains(
            file,
            "MACROSCOPIC STATIC DIELECTRIC TENSOR (including local field effects",
        )
    except NoMatchingLineFoundException as exc:
        raise InvalidFileException("polarizability not found") from exc
    file.readline()

    polarizability = []
    for _ in range(3):
        try:
            line = file.readline()
            polarizability.append([float(item) for item in line.split()[0:3]])
        except (EOFError, ValueError, IndexError) as exc:
            raise InvalidFileException(
                f"polarizability could not be parsed: {line}"
            ) from exc
    return np.array(polarizability)


def _get_lattice_vector_from_outcar_line(line: str) -> NDArray[np.float64]:
    """Read lattice vector from an appropriate line in a VASP OUTCAR file.

    Raises
    ------
    TypeError
    ValueError
    """
    try:
        return np.array([float(item) for item in line.split()[0:3]])
    except TypeError as exc:
        raise TypeError("line is not a str") from exc
    except IndexError as exc:
        raise ValueError("line does not have the expected format") from exc


def _read_lattice(file: TextIO) -> NDArray[np.float64]:
    """Read all three lattice vectors (in angstroms) from a VASP OUTCAR file.

    Raises
    ------
    InvalidFileException
    """
    # HACK: if symmetry is turned on, outcar will write additional
    # lattice vectors. We must skip ahead.
    try:
        _ = _skip_file_until_line_contains(
            file,
            "Write flags",
        )
        _ = _skip_file_until_line_contains(
            file,
            "direct lattice vectors      ",
        )
    except NoMatchingLineFoundException as exc:
        raise InvalidFileException("outcar does not have expected format") from exc

    lattice = []
    for _ in range(3):
        try:
            line = file.readline()
            lattice.append(_get_lattice_vector_from_outcar_line(line))
        except (EOFError, ValueError) as exc:
            raise InvalidFileException(f"lattice could not be parsed: {line}") from exc

    return np.array(lattice)


def read_phonons(filepath: str | Path) -> Phonons:
    """Read phonons from a VASP OUTCAR file.

    Parameters
    ----------
    filepath

    Returns
    -------
    :

    Raises
    ------
    FileNotFoundError
        File not found.
    InvalidFileException
        Invalid file.
    """
    wavenumbers = []
    eigenvectors = []

    filepath = pathify(filepath)
    with open(filepath, "r", encoding="utf-8") as file:

        # get atom information
        atomic_symbols = _read_atomic_symbols(file)
        atomic_weights = np.array([ATOMIC_WEIGHTS[symbol] for symbol in atomic_symbols])
        num_atoms = len(atomic_symbols)
        lattice = _read_lattice(file)
        ref_positions = _read_positions(file, len(atomic_symbols))
        num_degrees_of_freedom = num_atoms * 3

        # read in eigenvectors/eigenvalues
        try:
            _ = _skip_file_until_line_contains(
                file, "Eigenvectors and eigenvalues of the dynamical matrix"
            )
        except NoMatchingLineFoundException as exc:
            raise InvalidFileException(
                "eigenvector/eigenvalues block not found"
            ) from exc
        for _ in range(num_degrees_of_freedom):
            try:
                line = _skip_file_until_line_contains(file, "cm-1")
                if "f/i" in line:  # if complex
                    wavenumbers.append(
                        -float(line.split()[6])  # set negative wavenumber
                    )
                else:
                    wavenumbers.append(float(line.split()[7]))
            except (NoMatchingLineFoundException, TypeError, IndexError) as exc:
                raise InvalidFileException("eigenvalue could not be parsed") from exc
            eigenvectors.append(_read_eigenvector(file, num_atoms, lattice))

        # Divide eigenvectors by sqrt(mass) to get cartesian displacements
        wavenumbers = np.array(wavenumbers)
        eigenvectors = np.array(eigenvectors)
        displacements = eigenvectors / np.sqrt(atomic_weights)[:, np.newaxis]

        return Phonons(ref_positions, wavenumbers, displacements)


def read_positions_and_polarizability(
    filepath: str | Path,
) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
    """Read fractional positions and polarizability from a VASP OUTCAR file.

    The polarizability returned by VASP is, in fact, a dielectric tensor. However,
    this is inconsequential to the calculation of Raman spectra.

    Parameters
    ----------
    filepath

    Returns
    -------
    :
        0.  positions -- (fractional) Array with shape (N,3) where N is the number of
            atoms.
        #.  polarizability -- Array with shape (3,3).

    Raises
    ------
    FileNotFoundError
        File not found.
    InvalidFileException
        Invalid file.
    """
    filepath = pathify(filepath)
    with open(filepath, "r", encoding="utf-8") as outcar_file:
        num_atoms = len(_read_atomic_symbols(outcar_file))
        positions = _read_positions(outcar_file, num_atoms)
        polarizability = _read_polarizability(outcar_file)
        return positions, polarizability


def read_positions(filepath: str | Path) -> NDArray[np.float64]:
    """Read fractional positions from a VASP OUTCAR file.

    Parameters
    ----------
    filepath

    Returns
    -------
    :
        (fractional) Array with shape (N,3) where N is the number of atoms.

    Raises
    ------
    FileNotFoundError
        File not found.
    InvalidFileException
        Invalid file.
    """
    filepath = pathify(filepath)
    with open(filepath, "r", encoding="utf-8") as file:
        num_atoms = len(_read_atomic_symbols(file))
        positions = _read_positions(file, num_atoms)
        return positions


def read_structure_and_polarizability(
    filepath: str | Path,
) -> tuple[NDArray[np.float64], list[int], NDArray[np.float64], NDArray[np.float64]]:
    """Read lattice, fractional positions, atomic numbers, polarizability from OUTCAR.

    The polarizability returned by VASP is, in fact, a dielectric tensor. However,
    this is inconsequential to the calculation of Raman spectra.

    Parameters
    ----------
    filepath

    Returns
    -------
    :
        0.  lattice -- (Å) Array with shape (3,3).
        #.  atomic_numbers -- List of length N where N is the number of atoms.
        #.  positions -- (fractional) Array with shape (N,3) where N is the number of
            atoms.
        #.  polarizability -- Array with shape (3,3).

    Raises
    ------
    FileNotFoundError
        File not found.
    InvalidFileException
        Invalid file.
    """
    filepath = pathify(filepath)
    with open(filepath, "r", encoding="utf-8") as outcar_file:
        atomic_symbols = _read_atomic_symbols(outcar_file)
        atomic_numbers = [ATOMIC_NUMBERS[symbol] for symbol in atomic_symbols]
        lattice = _read_lattice(outcar_file)
        positions = _read_positions(outcar_file, len(atomic_numbers))
        polarizability = _read_polarizability(outcar_file)
        return lattice, atomic_numbers, positions, polarizability


def read_polarizability_dataset(
    filepaths: str | Path | list[str] | list[Path],
) -> "PolarizabilityDataset":
    """Read polarizability dataset from OUTCAR files.

    Parameters
    ----------
    filepaths

    Returns
    -------
    :

    Raises
    ------
    FileNotFoundError
        File not found.
    InvalidFileException
        Invalid file.
    IncompatibleFileException
        File is incompatible with the dataset.
    """
    return _read_polarizability_dataset(filepaths, read_structure_and_polarizability)


def read_ref_structure(filepath: str | Path) -> ReferenceStructure:
    """Read reference structure from a VASP OUTCAR file.

    If the file contains multiple structures (such as those generated by a molecular
    dynamics run), the initial structure will be considered the reference structure.

    Parameters
    ----------
    filepath

    Raises
    ------
    FileNotFoundError
        File not found.
    InvalidFileException
        Invalid file.
    SymmetryException
        Structural symmetry determination failed.
    """
    filepath = pathify(filepath)
    with open(filepath, "r", encoding="utf-8") as file:
        atomic_symbols = _read_atomic_symbols(file)
        atomic_numbers = [ATOMIC_NUMBERS[symbol] for symbol in atomic_symbols]
        lattice = _read_lattice(file)
        positions = _read_positions(file, len(atomic_symbols))
        return ReferenceStructure(atomic_numbers, lattice, positions)


def _read_next_cart_positions_ts(file: TextIO, num_atoms: int) -> NDArray[np.float64]:
    """Read Cartesian positions contained in a VASP OUTCAR file from an MD run.

    This function assumes that Cartesian positions are ready to be read.

    Raises
    ------
    InvalidFileException
    NoMatchingLineFoundException
    """
    file.readline()
    cart_positions = []
    for _ in range(num_atoms):
        try:
            line = file.readline()
            cart_positions.append([float(item) for item in line.split()[0:3]])
        except (EOFError, ValueError, IndexError) as exc:
            raise InvalidFileException(
                f"Cartesian positions could not be parsed: {line}"
            ) from exc
    return np.array(cart_positions)


def _read_timestep(file: TextIO) -> float:
    """Read timestep (in fs) contained in a VASP OUTCAR file.

    Raises
    ------
    InvalidFileException
    """
    try:
        line = _skip_file_until_line_contains(file, "time-step for ionic-motion")
    except NoMatchingLineFoundException as exc:
        raise InvalidFileException("timestep not found") from exc
    try:
        return float(line.split()[2])
    except (IndexError, TypeError) as exc:
        raise InvalidFileException(f"timestep could not be parsed: {line}") from exc


def read_trajectory(filepath: str | Path) -> Trajectory:
    """Read molecular dynamics trajectory from a VASP OUTCAR file.

    Parameters
    ----------
    filepath

    Raises
    ------
    FileNotFoundError
        File not found.
    InvalidFileException
        Invalid file.
    """
    filepath = pathify(filepath)
    with open(filepath, "r", encoding="utf-8") as file:
        atomic_symbols = _read_atomic_symbols(file)
        timestep = _read_timestep(file)
        lattice = _read_lattice(file)

        ml_step = False
        positions_ts = []
        while True:
            try:
                line = _skip_file_until_line_contains(
                    file,
                    "POSITION                                       TOTAL-FORCE ",
                )
            except NoMatchingLineFoundException:
                break

            if ml_step and "(ML)" not in line:
                ml_step = "(ML)" in line
                continue  # we skip ab initio following ML step
            ml_step = "(ML)" in line
            cart_positions = _read_next_cart_positions_ts(file, len(atomic_symbols))
            positions_ts.append(cart_positions @ np.linalg.inv(lattice))

        if len(positions_ts) == 0:
            raise InvalidFileException("no trajectory found")

        return Trajectory(np.array(positions_ts), timestep)



"""Raman spectrum classes and utilities."""

import numpy as np
from numpy.typing import NDArray

from ramannoodle.exceptions import get_type_error, verify_ndarray_shape
from ramannoodle.constants import BOLTZMANN_CONSTANT
from ramannoodle.abstract import RamanSpectrum

from ramannoodle.spectrum.utils import calc_signal_spectrum


def get_bose_einstein_correction(
    wavenumbers: NDArray[np.float64], temperature: float
) -> NDArray[np.float64]:
    r"""Calculate Bose-Einstein spectral correction.

    Parameters
    ----------
    wavenumbers
        (cm\ :sup:`-1`) Array with shape (M,).
    temperature
        (K)

    Returns
    -------
    :
        Array with shape (M,).

    """
    try:
        if temperature <= 0:
            raise ValueError(f"invalid temperature: {temperature} <= 0")
    except TypeError as exc:
        raise get_type_error("temperature", temperature, "float") from exc
    try:
        energy = wavenumbers * 29979245800.0 * 4.1357e-15  # in eV
        return 1 / (1 - np.exp(-energy / (BOLTZMANN_CONSTANT * temperature)))
    except TypeError as exc:
        raise get_type_error("wavenumbers", wavenumbers, "ndarray") from exc


def get_laser_correction(
    wavenumbers: NDArray[np.float64], laser_wavenumber: float
) -> NDArray[np.float64]:
    r"""Calculate conventional laser-wavenumber-dependent spectral correction.

    Parameters
    ----------
    wavenumbers
        (cm\ :sup:`-1`) Array with shape (M,).
    laser_wavenumber
        (cm\ :sup:`-1`)

    Returns
    -------
    :
        Array with shape (M,).

    """
    try:
        if laser_wavenumber <= 0:
            raise ValueError(f"invalid laser_wavenumber: {laser_wavenumber} <= 0")
    except TypeError as exc:
        raise get_type_error("laser_wavenumber", laser_wavenumber, "float") from exc
    try:
        return ((wavenumbers - laser_wavenumber) / 10000) ** 4 / wavenumbers
    except TypeError as exc:
        raise get_type_error("wavenumbers", wavenumbers, "ndarray") from exc


class PhononRamanSpectrum(RamanSpectrum):
    r"""Phonon-based first-order Raman spectrum.

    The spectrum is specified by a list of phonon wavenumbers and corresponding Raman
    tensors.

    Parameters
    ----------
    phonon_wavenumbers
        (cm\ :sup:`-1`) Array with shape (M,) where M is the number of phonons.
    raman_tensors
        Array with shape (M,3,3).

    """

    def __init__(
        self,
        phonon_wavenumbers: NDArray[np.float64],
        raman_tensors: NDArray[np.float64],
    ) -> None:
        verify_ndarray_shape("phonon_wavenumbers", phonon_wavenumbers, (None,))
        verify_ndarray_shape(
            "raman_tensors", raman_tensors, (len(phonon_wavenumbers), 3, 3)
        )
        self._phonon_wavenumbers = phonon_wavenumbers
        self._raman_tensors = raman_tensors

    @property
    def phonon_wavenumbers(self) -> NDArray[np.float64]:
        r"""Get (a copy of) phonon wavenumbers.

        Returns
        -------
        :
            (cm\ :sup:`-1`) Array with shape (M,) where M is the number of phonons.
        """
        return self._phonon_wavenumbers.copy()

    @property
    def raman_tensors(self) -> NDArray[np.float64]:
        """Get (a copy of) Raman tensors.

        Returns
        -------
        :
            Array with shape (M,3,3) where M is the number of phonons.
        """
        return self._raman_tensors.copy()

    # pylint: disable=too-many-arguments,too-many-positional-arguments
    def measure(
        self,
        orientation: str | NDArray[np.float64] = "polycrystalline",
        laser_correction: bool = False,
        laser_wavelength: float = 522,
        bose_einstein_correction: bool = False,
        temperature: float = 300,
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        r"""Calculate and return a raw Raman spectrum.

        Parameters
        ----------
        orientation
            Supports ``"polycrystalline"``. Future versions will support arbitrary
            orientations.
        laser_correction
            If ``True``, applies laser-wavelength-dependent intensity correction.
        laser_wavelength
            (nm) Ignored if ``laser_correction == False``.
        bose_einstein_correction
            If ``True``, applies temperature-dependent Bose Einstein correction.
        temperature
            (K) Ignored if ``bose_einstein_correction == False``.

        Returns
        -------
        :
            0.  wavenumbers -- (cm\ :sup:`-1`) Array with shape (M,).
            #.  intensities -- (arbitrary units) Array with shape (M,).

        Raises
        ------
        NotImplementedError
            Raised when any orientation besides ``"polycrystalline"`` is supplied.

        """
        if orientation != "polycrystalline":
            raise NotImplementedError(
                "only polycrystalline spectra are supported for now"
            )

        alpha_squared = (
            (
                self._raman_tensors[:, 0, 0]
                + self._raman_tensors[:, 1, 1]
                + self._raman_tensors[:, 2, 2]
            )
            / 3.0
        ) ** 2
        gamma_squared = (
            (self._raman_tensors[:, 0, 0] - self._raman_tensors[:, 1, 1]) ** 2
            + (self._raman_tensors[:, 0, 0] - self._raman_tensors[:, 2, 2]) ** 2
            + (self._raman_tensors[:, 1, 1] - self._raman_tensors[:, 2, 2]) ** 2
            + 6.0
            * (
                self._raman_tensors[:, 0, 1] ** 2
                + self._raman_tensors[:, 0, 2] ** 2
                + self._raman_tensors[:, 1, 2] ** 2
            )
        ) / 2.0
        intensities = 45.0 * alpha_squared + 7.0 * gamma_squared

        if laser_correction:
            laser_wavenumber = 10000000 / laser_wavelength
            intensities *= get_laser_correction(
                self._phonon_wavenumbers, laser_wavenumber
            )
        if bose_einstein_correction:
            intensities *= get_bose_einstein_correction(
                self._phonon_wavenumbers, temperature
            )

        return self._phonon_wavenumbers, intensities


class MDRamanSpectrum(RamanSpectrum):
    """Molecular-dynamics-based Raman spectrum.

    The spectrum is specified by a polarizability time series (``polarizability_ts``)
    and a timestep.

    Parameters
    ----------
    polarizability_ts
        Array with shape (S,3,3) where S is the number of configurations.
    timestep
        (fs)

    """

    def __init__(self, polarizability_ts: NDArray[np.float64], timestep: float):
        verify_ndarray_shape("polarizability_ts", polarizability_ts, (None, 3, 3))

        self._polarizability_ts = polarizability_ts
        self._timestep = timestep

    @property
    def polarizability_ts(self) -> NDArray[np.float64]:
        """Get (a copy) of polarizability time series.

        Returns
        -------
        :
            Array with shape (S,3,3) where S is the number of configurations.
        """
        return self._polarizability_ts

    @property
    def timestep(self) -> float:
        """Get timestep.

        Returns
        -------
        :
            (fs)
        """
        return self._timestep

    # pylint: disable=too-many-arguments,too-many-positional-arguments
    def measure(
        self,
        orientation: str | NDArray[np.float64] = "polycrystalline",
        laser_correction: bool = False,
        laser_wavelength: float = 522,
        bose_einstein_correction: bool = False,
        temperature: float = 300,
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        r"""Calculate and return a raw Raman spectrum.

        .. note:: Raw MD-derived Raman spectra will typically need to be smoothed to be
                  visualized effectively. See :func:`.spectrum_utils.convolve_spectrum`.

        Parameters
        ----------
        orientation
            Supports ``"polycrystalline"``. Future versions will support arbitrary
            orientations.
        laser_correction
            If ``True``, applies laser-wavelength-dependent intensity correction.
        laser_wavelength
            (nm) Ignored if ``laser_correction == False``.
        bose_einstein_correction
            If ``True``, applies temperature-dependent Bose Einstein correction.
        temperature
            (K) Ignored if ``bose_einstein_correction == False``.

        Returns
        -------
        :
            0.  wavenumbers -- (cm\ :sup:`-1`) Array with shape (ceiling(S / 2),) where
                S is the number of configurations.
            #.  intensities -- (arbitrary units) Array with shape (ceiling(S / 2),).

        """
        if orientation != "polycrystalline":
            raise NotImplementedError(
                "only polycrystalline spectra are supported for now"
            )

        # "alpha dot" aka polarizability time derivative
        ad = np.diff(self._polarizability_ts, axis=0)
        timestep = self._timestep
        wavenumbers, _ = calc_signal_spectrum(ad[:, 0, 0], timestep)

        alpha2 = (1 / 9) * calc_signal_spectrum(
            ad[:, 0, 0] + ad[:, 1, 1] + ad[:, 2, 2], timestep
        )[1]
        gamma2 = (
            (1 / 2) * calc_signal_spectrum(ad[:, 0, 0] - ad[:, 1, 1], timestep)[1]
            + (1 / 2) * calc_signal_spectrum(ad[:, 1, 1] - ad[:, 2, 2], timestep)[1]
            + (1 / 2) * calc_signal_spectrum(ad[:, 2, 2] - ad[:, 0, 0], timestep)[1]
            + 3 * calc_signal_spectrum(ad[:, 0, 1], timestep)[1]
            + 3 * calc_signal_spectrum(ad[:, 1, 2], timestep)[1]
            + 3 * calc_signal_spectrum(ad[:, 0, 2], timestep)[1]
        )
        intensities = 45.0 * alpha2 + 7.0 * gamma2

        # Intensity @ 0 cm-1 is infinite, so we remove.
        intensities = intensities[1:]
        wavenumbers = wavenumbers[1:]

        if laser_correction:
            laser_wavenumber = 10000000 / laser_wavelength
            intensities *= get_laser_correction(wavenumbers, laser_wavenumber)
        if bose_einstein_correction:
            intensities *= get_bose_einstein_correction(wavenumbers, temperature)

        return wavenumbers, intensities



"""Functions for interacting with VASP files."""

# flake8: noqa: F401
from ramannoodle.io.vasp import outcar
from ramannoodle.io.vasp import poscar
from ramannoodle.io.vasp import xdatcar
from ramannoodle.io.vasp import vasprun

__all__ = ["outcar", "poscar", "xdatcar", "vasprun"]



"""Classes and functions for calculating and manipulating spectra."""

# flake8: noqa: F401
from ramannoodle.spectrum._raman import (
    PhononRamanSpectrum,
    MDRamanSpectrum,
)
from ramannoodle.spectrum import utils

__all__ = ["PhononRamanSpectrum", "MDRamanSpectrum", "utils"]



"""Training functions for PotGNN polarizability model."""

import numpy as np
from numpy.typing import NDArray

from ramannoodle.exceptions import (  # pylint: disable=ungrouped-imports
    get_torch_missing_error,
    UserError,
)

try:
    import torch
    from torch.utils.data import DataLoader
    from torch.nn.modules.loss import _Loss
    from torch.optim.optimizer import Optimizer

    from ramannoodle.pmodel.torch._gnn import PotGNN
    from ramannoodle.dataset.torch._dataset import PolarizabilityDataset
except (ModuleNotFoundError, UserError) as exc:
    raise get_torch_missing_error() from exc


# pylint: disable=too-many-arguments,too-many-positional-arguments,too-many-locals
def train_single_epoch(
    model: PotGNN,
    training_set: PolarizabilityDataset,
    validation_set: PolarizabilityDataset,
    batch_size: int,
    optimizer: Optimizer,
    loss_function: _Loss,
) -> tuple[float, float, NDArray[np.float64]]:
    """Train PotGNN model for a single epoch on the default device.

    Parameters
    ----------
    model
    training_set
    validation_set
    batch_size
    optimizer
    loss_function

    Returns
    -------
    :
        0.  mean training loss
        #.  mean validation loss
        #.  mean variance of predictions on validation set -- Array with shape [6,]

    """
    default_device = torch.get_default_device()

    train_loader = DataLoader(
        training_set,
        batch_size=batch_size,
        shuffle=True,
        generator=torch.Generator(device=default_device),
    )
    validation_loader = DataLoader(
        validation_set, batch_size=min(100, len(validation_set)), shuffle=False
    )

    model.train()
    train_losses = []
    for lattice, atomic_numbers, position, polarizability in train_loader:
        out = model.forward(
            lattice.to(default_device),
            atomic_numbers.to(default_device),
            position.to(default_device),
        )
        loss = loss_function(out, polarizability)
        train_losses.append(float(loss))

        loss.backward()
        optimizer.step()
        optimizer.zero_grad()

    model.eval()
    validation_losses = []
    validation_vars = []
    for lattice, atomic_numbers, position, polarizability in validation_loader:
        out = model.forward(
            lattice.to(default_device),
            atomic_numbers.to(default_device),
            position.to(default_device),
        )
        loss = loss_function(out, polarizability)
        validation_losses.append(float(loss))
        validation_vars.append(torch.var(out, dim=0).detach().cpu().numpy().copy())

    return (
        float(np.mean(train_losses)),
        float(np.mean(validation_losses)),
        np.mean(validation_vars, axis=0),
    )



"""Some torch utilities."""

from typing import Sequence

from ramannoodle.exceptions import (
    get_type_error,
    get_torch_missing_error,
)

try:
    import torch
    from torch import Tensor
except ModuleNotFoundError as exc:
    raise get_torch_missing_error() from exc

# pylint complains about torch.norm
# pylint: disable=not-callable


def polarizability_vectors_to_tensors(polarizability_vectors: Tensor) -> Tensor:
    """Convert polarizability vectors to symmetric tensors.

    Parameters
    ----------
    polarizability_vectors
        Tensor with size [S,6].

    Returns
    -------
    :
        Tensor with size [S,3,3].
    """
    verify_tensor_size("polarizability_vectors", polarizability_vectors, (None, 6))
    indices = torch.tensor(
        [
            [0, 3, 4],
            [3, 1, 5],
            [4, 5, 2],
        ]
    )
    return polarizability_vectors[:, indices]


def polarizability_tensors_to_vectors(polarizability_tensors: Tensor) -> Tensor:
    """Convert polarizability tensors to vectors.

    Parameters
    ----------
    polarizability_tensors
        Tensor with size [S,3,3] where S is the number of samples.

    Returns
    -------
    :
        Tensor with size [S,6].

    """
    verify_tensor_size("polarizability_tensors", polarizability_tensors, (None, 3, 3))
    indices = torch.tensor([[0, 0], [1, 1], [2, 2], [0, 1], [0, 2], [1, 2]]).T
    return polarizability_tensors[:, indices[0], indices[1]]


def _get_tensor_size_str(size: Sequence[int | None]) -> str:
    """Get a string representing a tensor size.

    "_" indicates a dimension can be any size.

    Parameters
    ----------
    size
        None indicates dimension can be any size.
    """
    result = "["
    for i in size:
        if i is None:
            result += "_,"
        else:
            result += f"{i},"
    if len(size) == 1:
        return result + "]"
    return result[:-1] + "]"


def get_tensor_size_error(name: str, tensor: Tensor, desired_size: str) -> ValueError:
    """Get ValueError indicating a PyTorch Tensor has the wrong size."""
    try:
        shape_spec = f"{_get_tensor_size_str(tensor.size())} != {desired_size}"
    except AttributeError as exc:
        raise get_type_error("tensor", tensor, "Tensor") from exc
    return ValueError(f"{name} has wrong size: {shape_spec}")


def verify_tensor_size(name: str, tensor: Tensor, size: Sequence[int | None]) -> None:
    """Verify a PyTorch Tensor's size.

    :meta private: We should avoid calling this function whenever possible (EATF).

    Parameters
    ----------
    size
        int elements will be checked, None elements will not be.
    """
    try:
        if len(size) != tensor.ndim:
            raise get_tensor_size_error(name, tensor, _get_tensor_size_str(size))
        for d1, d2 in zip(tensor.size(), size, strict=True):
            if d2 is not None and d1 != d2:
                raise get_tensor_size_error(name, tensor, _get_tensor_size_str(size))
    except AttributeError as exc:
        raise get_type_error(name, tensor, "Tensor") from exc



"""Universal IO utility functions."""

from typing import TextIO, Callable
from pathlib import Path

import numpy as np
from numpy.typing import NDArray
from tqdm import tqdm

from ramannoodle.exceptions import (
    NoMatchingLineFoundException,
    verify_ndarray_shape,
    verify_positions,
    verify_list_len,
    IncompatibleStructureException,
    get_torch_missing_error,
    UserError,
)
from ramannoodle.constants import ATOM_SYMBOLS

TORCH_PRESENT = True
try:
    from ramannoodle.dataset.torch._dataset import PolarizabilityDataset
except UserError:
    TORCH_PRESENT = False


def _skip_file_until_line_contains(file: TextIO, content: str) -> str:
    """Read through a file until a line containing specific content is found."""
    for line in file:
        if content in line:
            return line
    raise NoMatchingLineFoundException(content)


def pathify(filepath: str | Path) -> Path:
    """Convert filepath to Path.

    :meta private:
    """
    try:
        return Path(filepath)
    except TypeError as exc:
        raise TypeError(f"{filepath} cannot be resolved as a filepath") from exc


def pathify_as_list(filepaths: str | Path | list[str] | list[Path]) -> list[Path]:
    """Convert filepaths to list of Paths.

    :meta private:
    """
    if isinstance(filepaths, list):
        paths = []
        if len(filepaths) == 0:
            raise ValueError("filepaths is empty")
        for item in filepaths:
            paths.append(pathify(item))
        return paths
    return [pathify(filepaths)]


def verify_structure(
    lattice: NDArray[np.float64],
    atomic_numbers: list[int],
    positions: NDArray[np.float64],
) -> None:
    """Verify a structure.

    :meta private:
    """
    verify_ndarray_shape("lattice", lattice, (3, 3))
    verify_list_len("atomic_numbers", atomic_numbers, None)
    for atomic_number in atomic_numbers:
        if atomic_number not in ATOM_SYMBOLS.keys():
            raise ValueError(f"invalid atomic number: {atomic_number}")
    verify_ndarray_shape("positions", positions, (len(atomic_numbers), 3))
    verify_positions("positions", positions)


def verify_trajectory(
    lattice: NDArray[np.float64],
    atomic_numbers: list[int],
    positions_ts: NDArray[np.float64],
) -> None:
    """Verify a trajectory.

    :meta private:
    """
    verify_ndarray_shape("lattice", lattice, (3, 3))
    verify_list_len("atomic_numbers", atomic_numbers, None)
    for atomic_number in atomic_numbers:
        if atomic_number not in ATOM_SYMBOLS.keys():
            raise ValueError(f"invalid atomic number: {atomic_number}")
    verify_ndarray_shape("positions_ts", positions_ts, (None, len(atomic_numbers), 3))
    if (0 > positions_ts).any() or (positions_ts > 1.0).any():
        raise ValueError("positions_ts has coordinates that are not between 0 and 1")


def _read_polarizability_dataset(
    filepaths: str | Path | list[str] | list[Path],
    read_structure_and_polarizability_fn: Callable[
        [str | Path],
        tuple[NDArray[np.float64], list[int], NDArray[np.float64], NDArray[np.float64]],
    ],
) -> "PolarizabilityDataset":
    """Read polarizability dataset from OUTCAR files.

    Parameters
    ----------
    filepath
    read_structure_and_polarizability_fn

    Returns
    -------
    :

    Raises
    ------
    FileNotFoundError
    InvalidFileException
        Invalid file.
    IncompatibleFileException
        File is incompatible with the dataset.
    """
    if not TORCH_PRESENT:
        raise get_torch_missing_error()
    filepaths = pathify_as_list(filepaths)

    lattice = np.zeros((3, 3))
    atomic_numbers: list[int] = []
    positions_list: list[NDArray[np.float64]] = []
    polarizabilities: list[NDArray[np.float64]] = []
    for file_index, filepath in tqdm(list(enumerate(filepaths)), unit=" files"):
        read_lattice, read_atomic_numbers, positions, polarizability = (
            read_structure_and_polarizability_fn(filepath)
        )
        if file_index != 0:
            if not np.allclose(lattice, read_lattice, atol=1e-5):
                raise IncompatibleStructureException(
                    f"incompatible lattice: {filepath}"
                )
            if atomic_numbers != read_atomic_numbers:
                raise IncompatibleStructureException(
                    f"incompatible atomic numbers: {filepath}"
                )
            if positions_list[0].shape != positions.shape:  # check, just to be safe
                raise IncompatibleStructureException(
                    f"incompatible atomic positions: {filepath}"
                )
        lattice = read_lattice
        atomic_numbers = read_atomic_numbers
        positions_list.append(positions)
        polarizabilities.append(polarizability)

    return PolarizabilityDataset(
        lattice,
        atomic_numbers,
        np.array(positions_list),
        np.array(polarizabilities),
    )



"""Classes and functions for atomic structures and structural symmetries."""

# flake8: noqa: F401
from ramannoodle.structure._reference import ReferenceStructure
from ramannoodle.structure._displace import (
    write_ast_displaced_structures,
    get_ast_displaced_positions,
    write_displaced_structures,
    get_displaced_positions,
)
from ramannoodle.structure import _symmetry_utils
from ramannoodle.structure import utils


__all__ = [
    "ReferenceStructure",
    "write_ast_displaced_structures",
    "get_ast_displaced_positions",
    "write_displaced_structures",
    "get_displaced_positions",
    "_symmetry_utils",
    "utils",
]



"""Facilitates calculation of Raman spectra from first principles calculations."""

# flake8: noqa: F401
from ramannoodle import exceptions
from ramannoodle import constants
from ramannoodle import spectrum
from ramannoodle import io
from ramannoodle import structure
from ramannoodle import pmodel
from ramannoodle import dynamics
from ramannoodle import dataset
from ramannoodle import abstract



"""Classes for various polarizability models."""

# flake8: noqa: F401
from ramannoodle.pmodel._art import ARTModel
from ramannoodle.pmodel._interpolation import InterpolationModel

__all__ = ["ARTModel", "InterpolationModel"]
